import gym
import pandas as pd
import numpy as np
from gym import spaces
import torch
from s_network import train

class MerlinEnv(gym.Env):
      """Custom Environment for the Merlin project based on historical data. The environment follows the gym interface as described in https://github.com/openai/gym/blob/master/docs/creating-environments.md"""
      # We need to initialize the metadata with render modes. Otherwise we will have an exception.
      # Since we are working on Databricks, we should not render anything.
      metadata = {'render.modes':['human']}
  
      def __init__(self, df, task, s_network, seq_len=5):
        """We load the pandas Dataframe with the data. Based on this dataframe, we will
        provide the states/actions at each step.
        The task can be Engagement, Adoption"""
        
        super(MerlinEnv, self).__init__()
        self.df = df.sort_values(by='timestamp')
        self.df = self.df.dropna()
        self.df['day_time_idx'] = 24 * self.df['timestamp'].dt.dayofweek + self.df['timestamp'].dt.hour
        self.task = task
        self.timezones = len(self.df.iloc[0]['timezones'])
        # Action as an one-hot vector with size 24 
        self.action_space = spaces.Discrete(24)
        
        # [num_viewers, engagement_score, adoption_score, timezones]
        high = np.concatenate((np.array([np.finfo(np.float16).max, 1., 1., 1.], dtype=np.float16),np.ones(self.timezones, dtype=np.float16)))

        self.observation_space = spaces.Box(-high, high, dtype=np.float16)
        
        self.seq_len = seq_len
        self.s_network = self.train_s_network()

    
  

      def train_s_network(self):
            return self.s_network.train(self.df)

      def _get_observations(self, idx=0):
        obs = None
        if idx + self.seq_len < self.df.shape[0]:
          data = self.df.iloc[idx: idx+self.seq_len]
          observations = np.array([np.concatenate(([row['num_viewers'], row['Engagement'], row['Adoption']/(idx+1)],row['timezones'])) for _, row in data.iterrows()], dtype=np.float32)
          obs = torch.unsqueeze(torch.tensor(observations),1)
        return obs
      
      
      def _get_reward(self,idx=0):
        reward = self.s_network(self._get_observations(idx))
        return reward
      
      def _get_next_action(self,idx=0):
        next_action=None
        if idx + self.seq_len < self.df.shape[0]:
          next_action = torch.tensor(self.df.iloc[idx+self.seq_len]['day_time_idx'])
        return next_action
      

      def _get_step_details(self,idx):
        current_observations = self._get_observations(idx)
        action = self._get_next_action(idx)
        next_observations = self._get_observations(idx + 1)
        reward = self._get_reward(idx)
        done = idx + self.seq_len >= self.df.shape[0] - 1
        return current_observations, action, next_observations, reward, done
      
      def step(self):    
        self.current_step += 1
        return self._get_step_details(self.current_step)
      
      def reset(self):
        self.current_step = 0
        return self._get_step_details(self.current_step)
      
      def render(self, mode='human'):
        pass
      
      def close(self):
        pass