import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torch.distributions as distributions

class State(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers=1, dropout=0.5, bidirectional=False):
        super().__init__()
        self.lstm = nn.LSTM(input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, dropout=dropout, bidirectional=bidirectional)
    
    def forward(self, x):
            output, (hn, cn) = self.lstm(x)
            return output[-1,:,:]
  
class MLP(nn.Module):
  def __init__(self, input_size, hidden_size, output_size, dropout=.5):
    super().__init__()
    self.fc_1 = nn.Linear(input_size, hidden_size)
    self.fc_2 = nn.Linear(hidden_size, output_size)
    self.dropout = nn.Dropout(dropout)
  
  def forward(self, x):
    x = self.fc_1(x)
    x = self.dropout(x)
    x = F.relu(x)
    x = self.fc_2(x)
    return x
    

class Actor(nn.Module):
  def __init__(self, input_size, hidden_size, output_size, num_layers=1, dropout=0.5, bidirectional=False):
    super().__init__()
    self.action = MLP(hidden_size, hidden_size, output_size, dropout)
  
  def forward(self, state):    
    return self.action(state)
  

class Critic(nn.Module):
  def __init__(self, state_input_size, action_input_size, hidden_size, output_size, dropout=.5):
    super().__init__()
    self.state_mlp = MLP(state_input_size, hidden_size, hidden_size, dropout)
    self.action_mlp = MLP(action_input_size, hidden_size, hidden_size, dropout)
    self.q_value = MLP(hidden_size, hidden_size, output_size, dropout)
    
    
  def forward(self, state, actions):
    state_vector = self.state_mlp(state)
    action_vector = self.action_mlp(actions)
    state_action_vector = torch.sum(torch.stack([state_vector, action_vector]),dim=0)
    q = self.q_value(state_action_vector)
    return q
    
def init_weights(m):
  if type(m) ==nn.Linear:
    torch.nn.init.xavier_normal_(m.weight)
    m.bias.data.fill_(0)


class ActorCritic(nn.Module):
  def __init__(self, actor, critic):
    super(ActorCritic, self).__init__()
    self.actor = actor
    self.critic = critic

  def forward(self, state):
    action_pred = self.actor(state)
    value_pred = self.critic(state)
    return action_pred, value_pred